import meshio
import logging

import pandas as pd
import numpy as np

logger = logging.getLogger(name=__name__)


def save_triangular_mesh(filename, eptm):
    coords = eptm.coords
    eptm.reset_index(order=True)

    if (filename[-3:] == 'ply') or (filename[-3:] == 'obj'):
        points, faces = eptm.triangular_mesh(coords=coords, )
        cells = []
        for f in faces:
            cells.append(("triangle", np.array([f])))
        mesh = meshio.Mesh(points, cells)
        mesh.write(filename)
    elif (filename[-3:] == 'vtk'):
        points, faces = eptm.vertex_mesh(coords=coords, vertex_normals=False)
        cells = []
        for f in faces:
            cells.append(("triangle", np.array([f])))
        mesh = meshio.Mesh(points, cells)
        meshio.vtk.write(filename, mesh)
    else:
        print("This format %s is not taking in charge for now", filename[-3:])

    logger.info("Saved %s as a meshio file", filename)


def import_triangular_mesh(filename):
    if (filename[-3:] == 'ply') or (filename[-3:] == 'obj'):
        mesh = meshio.read(filename)
        vert_ = pd.DataFrame(mesh.points, columns=list('xyz'))
        edge_ = pd.DataFrame(columns=['srce', 'trgt', 'face'])
        face_ = pd.DataFrame(columns=list('xyz'))
        cpt = 0
        for c in mesh.cells[0][1]:
            edge_.loc[cpt * 3] = [c[0], c[1], cpt]
            edge_.loc[cpt * 3 + 1] = [c[1], c[2], cpt]
            edge_.loc[cpt * 3 + 2] = [c[2], c[0], cpt]

            face_.loc[cpt] = [0, 0, 0]
            cpt += 1
        data = {'vert': vert_,
                'edge': edge_,
                'face': face_}

        return data
    else:
        print("This format %s is not taking in charge for now", filename[-3:])


def save_mesh(filename, eptm):
    """
    Saving mesh which is not triangular, need to call specifically the write
    function according to the format and not the generic write function
    """
    coords = eptm.coords
    eptm.reset_index(order=True)

    points, faces = eptm.vertex_mesh(coords=coords, vertex_normals=False)
    cells = []
    for f in faces:
        cells.append(("triangle", np.array([f])))
    mesh = meshio.Mesh(points, cells)

    if (filename[-3:] == 'ply'):
        meshio.ply.write(filename, mesh)
    elif (filename[-3:] == 'vtk'):
        meshio.vtk.write(filename, mesh)
    else:
        print("This format %s is not taking in charge for now", filename[-3:])

    logger.info("Saved %s as a meshio file", filename)


def import_mesh(filename):
    """
    Can only import ply file as polygonal sheet.
    """
    if (filename[-3:] == 'ply'):
        mesh = meshio.read(filename)
        vert_ = pd.DataFrame(mesh.points, columns=list('xyz'))
        edge_ = pd.DataFrame(columns=['srce', 'trgt', 'face'])
        face_ = pd.DataFrame(columns=['x', 'y', 'z', 'num_sides'])
        cpt_edge = 0
        cpt_face = 0

        for cells in mesh.cells:
            for c in cells.data:
                poly = len(c)
                for i in range(poly - 1):
                    edge_.loc[cpt_edge] = [c[i], c[i + 1], cpt_face]
                    cpt_edge += 1
                edge_.loc[cpt_edge] = [c[poly - 1], c[0], cpt_face]
                cpt_edge += 1
                face_.loc[cpt_face] = [0, 0, 0, poly]
                cpt_face += 1

        data = {'vert': vert_,
                'edge': edge_,
                'face': face_}

        return data
    else:
        print("This format %s is not taking in charge for now", filename[-3:])



import logging
import numpy as np

from ..geometry.bulk_geometry import MonolayerGeometry
from ..core.sheet import Sheet
from ..geometry.sheet_geometry import SheetGeometry
from ..geometry.utils import rotation_matrix

from .bulk_topology import get_division_vertices
from .bulk_topology import cell_division as bulk_division
from .sheet_topology import type1_transition as sheet_t1
from .sheet_topology import get_division_edges as sheet_division_edges


logger = logging.getLogger(name=__name__)


def cell_division(monolayer, mother, orientation="vertical", psi=None):
    """
    Divides the cell mother in the monolayer.

    Parameters
    ----------
    * monolayer: a :class:`Monolayer` instance
    * mother: int, the index of the cell to devide
    * orientation: str, {"vertical" | "horizontal" | "apical"}
      if "horizontal", performs a division in the equatorial
      plane of the cell. If "vertical" (the default), performs
      a division along the basal-apical axis of the cell.
      If "apical", performs a division cutting the apical face
      perpendicularly to its principal axis
    * psi: float, default 0
      extra rotation angle of the division plane
      around the basal-apical plane

    Returns
    -------
    * daughter: int, the index of the daughter cell
    """

    ab_axis = MonolayerGeometry.basal_apical_axis(monolayer, mother)

    if orientation == "horizontal":
        plane_normal = np.asarray(ab_axis)
    elif orientation == "vertical":
        plane_normal = _vertical_plane_normal(ab_axis, psi=psi)
    elif orientation == "apical":
        rcoords = ['r'+c for c in monolayer.coords]
        apical_pos = monolayer.edge_df.loc[
            (monolayer.edge_df['cell'] == mother)
            & (monolayer.edge_df['segment'] == "apical"),
            rcoords
        ]
        _, _, vh = np.linalg.svd(apical_pos)
        plane_normal = vh[0, :]

    else:
        raise ValueError(
            f"""orientation argument not understood, should be either "horizontal",
"vertical" or "apical", not {orientation}"""
        )

    vertices, mother_verts, daughter_verts = get_division_vertices(
        monolayer, mother=mother, plane_normal=plane_normal, return_all=True
    )
    daughter = bulk_division(
        monolayer, mother, MonolayerGeometry, vertices, mother_verts, daughter_verts
    )

    # Correct segment assignations for the septum
    septum = monolayer.face_df.index[-2:]
    septum_edges = monolayer.edge_df.index[-2 * len(vertices) :]
    if orientation == "vertical":
        monolayer.face_df.loc[septum, "segment"] = "lateral"
        monolayer.edge_df.loc[septum_edges, "segment"] = "lateral"
        _assign_vert_segment(monolayer, vertices)

    elif orientation == "horizontal":
        monolayer.face_df.loc[septum[0], "segment"] = "apical"
        monolayer.face_df.loc[septum[1], "segment"] = "basal"
        monolayer.edge_df.loc[septum_edges[: len(vertices)], "segment"] = "apical"
        monolayer.edge_df.loc[septum_edges[len(vertices) :], "segment"] = "basal"
        monolayer.vert_df.loc[vertices, "segment"] = "apical"

    return daughter


def _vertical_plane_normal(ab_axis, psi=None):

    # Find the simplest vector perpendicular to the ab_axis
    perp_axis = np.array([-ab_axis[1], ab_axis[0], 0])

    if psi is None:
        psi = np.random.uniform(0, np.pi)

    # rotate of an arbitrary angle
    return np.dot(rotation_matrix(psi, ab_axis), perp_axis)


def _assign_vert_segment(monolayer, vertices):

    for v in vertices:
        segs = set(monolayer.edge_df[monolayer.edge_df["srce"] == v]["segment"])
        if "apical" in segs:
            monolayer.vert_df.loc[v, "segment"] = "apical"
        elif "basal" in segs:
            monolayer.vert_df.loc[v, "segment"] = "basal"
        else:
            monolayer.vert_df.loc[v, "segment"] = "lateral"


def find_basal_edge(monolayer, apical_edge):
    """Returns the basal edge parallel to the apical edge passed
    in argument.

    Parameters
    ----------
    monolayer: a :class:`Monolayer` instance

    """
    srce, trgt, cell = monolayer.edge_df.loc[apical_edge, ["srce", "trgt", "cell"]]
    cell_edges = monolayer.edge_df[monolayer.edge_df["cell"] == cell]
    srce_segment = monolayer.vert_df.loc[cell_edges["srce"].values, "segment"]
    srce_segment.index = cell_edges.index
    trgt_segment = monolayer.vert_df.loc[cell_edges["trgt"].values, "segment"]
    trgt_segment.index = cell_edges.index
    try:
        (b_trgt,) = cell_edges[
            (srce_segment == "apical")
            & (trgt_segment == "basal")
            & (cell_edges["srce"] == srce)
        ]["trgt"]
        (b_srce,) = cell_edges[
            (srce_segment == "basal")
            & (trgt_segment == "apical")
            & (cell_edges["trgt"] == trgt)
        ]["srce"]
        (b_edge,) = cell_edges[
            (cell_edges["srce"] == b_srce) & (cell_edges["trgt"] == b_trgt)
        ].index
    except ValueError:
        b_edge = None

    return b_edge



import numpy as np
from .base_geometry import BaseGeometry


class PlanarGeometry(BaseGeometry):
    """Geomtetry methods for 2D planar cell arangements
    """

    @classmethod
    def update_all(cls, sheet):
        """
        Updates the sheet geometry by updating:
        * the edge vector coordinates
        * the edge lengths
        * the face centroids
        * the normals to each edge associated face
        * the face areas
        """

        cls.update_dcoords(sheet)
        cls.update_ucoords(sheet)
        cls.update_length(sheet)
        cls.update_centroid(sheet)
        cls.update_normals(sheet)
        cls.update_areas(sheet)
        cls.update_perimeters(sheet)

    @staticmethod
    def update_normals(sheet):

        rcoords = ["r" + c for c in sheet.coords]
        dcoords = ["d" + c for c in sheet.coords]

        normals = np.cross(sheet.edge_df[rcoords], sheet.edge_df[dcoords])
        sheet.edge_df["nz"] = normals

    @staticmethod
    def update_areas(sheet):
        """
        Updates the normal coordinate of each (srce, trgt, face) face.
        """
        sheet.edge_df["sub_area"] = sheet.edge_df["nz"] / 2
        sheet.face_df["area"] = sheet.sum_face(sheet.edge_df["sub_area"])

    @staticmethod
    def face_projected_pos(sheet, face, psi):
        """
        returns the sheet vertices position translated to center the face
        `face` at (0, 0) and rotated in the (x, y) plane
        by and angle `psi` radians

        """
        rot_pos = sheet.vert_df[sheet.coords].copy()
        face_x, face_y = sheet.face_df.loc[face, ["x", "y"]]
        rot_pos.x = (sheet.vert_df.x - face_x) * np.cos(psi) - (
            sheet.vert_df.y - face_y
        ) * np.sin(psi)
        rot_pos.y = (sheet.vert_df.x - face_x) * np.sin(psi) + (
            sheet.vert_df.y - face_y
        ) * np.cos(psi)

        return rot_pos

    @classmethod
    def get_phis(cls, sheet):
        if not "rx" in sheet.edge_df:
            cls.update_dcoords(sheet)
            cls.update_centroid(sheet)

        return np.arctan2(sheet.edge_df["ry"], sheet.edge_df["rx"])


# The following classes will probably be included in tyssue at some point
class AnnularGeometry(PlanarGeometry):
    """
    """

    @classmethod
    def update_all(cls, eptm):
        PlanarGeometry.update_all(eptm)
        cls.update_lumen_volume(eptm)

    @staticmethod
    def update_lumen_volume(eptm):
        srce_pos = eptm.upcast_srce(eptm.vert_df[["x", "y"]]).loc[
            eptm.apical_edges]
        trgt_pos = eptm.upcast_trgt(eptm.vert_df[["x", "y"]]).loc[
            eptm.apical_edges]
        apical_edge_pos = (srce_pos + trgt_pos) / 2
        apical_edge_coords = eptm.edge_df.loc[eptm.apical_edges, ["dx", "dy"]]
        eptm.settings["lumen_volume"] = (
            -apical_edge_pos["x"] * apical_edge_coords["dy"]
            + apical_edge_pos["y"] * apical_edge_coords["dx"]
        ).values.sum()


class WeightedPerimeterPlanarGeometry(PlanarGeometry):
    """
    Geometry methods for 2D planar cell arangements with a calculation
    of perimeter is based on weight of each junction.

    Meaning if all junction of a cell have the same weight, perimeter is
    calculated as a usual perimeter calculation (p = l_ij + l_jk + l_km + l_mn + l_ni)
    Otherwise, weight parameter allowed more or less importance of a junction in the
    perimeter calculation (p = w_ij*l_ij + w_jk*l_jk + w_km*l_km + w_mn*l_mn + w_ni*l_ni)

    """

    @classmethod
    def update_all(cls, eptm):
        cls.center(eptm)
        cls.normalize_weights(eptm)
        super().update_all(eptm)

    @staticmethod
    def update_perimeters(eptm):
        """
        Updates the perimeter of each face according to the weight of each junction.
        """
        eptm.edge_df["weighted_length"] = eptm.edge_df.weight * \
            eptm.edge_df.length
        eptm.face_df["perimeter"] = eptm.sum_face(
            eptm.edge_df["weighted_length"])

    @staticmethod
    def normalize_weights(sheet):
        """
        Normalize weight of each cell.
        Sum of all weights of one cell equals to one.
        """
        sheet.edge_df["num_sides"] = sheet.upcast_face("num_sides")
        sheet.edge_df["weight"] = (
            sheet.edge_df.groupby("face")
            .apply(lambda df: (df["num_sides"] * df["weight"] / df["weight"].sum()))
            .sort_index(level=1)
            .to_numpy()
        )
